@@ -8,9 +8,9 @@
 #'     then runs a series of generations of ABC-SMC. The generations can either be 
 #'     specified with a set of fixed tolerances, or by setting the tolerances at 
 #'     each new generation as a quantile of the tolerances of the accepted particles 
-#'     at the previous generation. Passing an \code{ABCSMC} object into the 
-#'     \code{ABCSMC()} function acts as a continuation method, allowing further 
-#'     generations to be run.
+#'     at the previous generation. Uses bisection method as detailed in McKinley et al. (2018).
+#'     Passing an \code{ABCSMC} object into the \code{ABCSMC()} function acts as a 
+#'     continuation method, allowing further generations to be run.
 #'
 #' @export
 #'
@@ -36,6 +36,7 @@
 #'                  The columns/entries must match to those in `x`. 
 #' @param ptols     The proportion of simulated outcomes at each generation to use to derive adaptive 
 #'                  tolerances.
+#' @param mintols   A vector of minimum tolerance levels.
 #' @param ngen      The number of generations of ABC-SMC to run.
 #' @param npart     An integer specifying the number of particles.
 #' @param parallel  A \code{logical} determining whether to use parallel processing or not.
@@ -63,6 +64,7 @@
 #'                      acceptance rates for each generation of ABC;}
 #' \item{\code{tols}:}{ a copy of the \code{tols} input;}
 #' \item{\code{ptols}:}{ a copy of the \code{ptols} input;}
+#' \item{\code{mintols}:}{ a copy of the \code{mintols} input;}
 #' \item{\code{priors}:}{ a copy of the \code{priors} input;}
 #' \item{\code{data}:}{ a copy of the \code{data} input;}
 #' \item{\code{func}:}{ a copy of the \code{func} input;}
@@ -74,6 +76,8 @@
 #' @references Toni T, Welch D, Strelkowa N, Ipsen A and Stumpf MP (2009) <doi:10.1098/rsif.2008.0172>
 #'     
 #' @references McKinley TJ, Cook AR and Deardon R (2009) <doi:10.2202/1557-4679.1171>
+#' 
+#' @references McKinley TJ, Vernon I, Andrianakis I, McCreesh N, Oakley JE, Nsubuga RN, Goldstein M and White RG (2018) <doi:10.1214/17-STS618>
 #'     
 #' @seealso \code{\link{print.ABCSMC}}, \code{\link{plot.ABCSMC}}, \code{\link{summary.ABCSMC}}
 #'     
@@ -181,7 +185,7 @@ ABCSMC <- function(x, ...) {
 #' @rdname ABCSMC
 #' @export
 
-ABCSMC.ABCSMC <- function(x, tols = NULL, ptols = NULL, ngen = 1, parallel = FALSE, mc.cores = NA, ...) {
+ABCSMC.ABCSMC <- function(x, tols = NULL, ptols = NULL, mintols = NULL, ngen = 1, parallel = FALSE, mc.cores = NA, ...) {
     
     ## check inputs
     if(class(x) != "ABCSMC"){
@@ -194,6 +198,11 @@ ABCSMC.ABCSMC <- function(x, tols = NULL, ptols = NULL, ngen = 1, parallel = FAL
         stop("Must choose either 'tols' or 'ptols'")
     }
     
+    ## set ne mintols if needed
+    if(!is.null(mintols)) {
+        x$mintols <- mintols
+    }
+    
     if(!is.null(tols[1])){
         ## extract tolerances and check against new tolerances
         if(!is.matrix(tols) & !is.vector(tols)){
@@ -212,15 +221,36 @@ ABCSMC.ABCSMC <- function(x, tols = NULL, ptols = NULL, ngen = 1, parallel = FAL
         }
         checkInput(tols, c("numeric"))
         if(sum(apply(rbind(x$tols[nrow(x$tols), ], tols), 2, function(x) {
-            sum(x[-1] >= x[1])
+            sum(x[-1] > x[1])
         })) > 0) {
-            stop("New tolerances not less than original tolerances")
+            stop("New tolerances not less than or equal to original tolerances")
+        }
+        ngen <- ifelse(is.null(nrow(tols)), 1, nrow(tols))
+        temp <- rbind(x$tols[nrow(x$tols), ], tols)
+        if(!all(apply(temp, 2, function(x) {
+            all(diff(x) <= 0)
+        }))){
+            stop("'tols' cannot increase")
+        }
+        temp <- apply(temp, 2, function(x) {
+            diff(x)
+        })
+        if(is.null(nrow(temp))) {
+            temp <- matrix(temp, nrow = 1)
+        }
+        if(any(apply(temp, 1, function(x) !any(x < 0)))){
+            stop("Some 'tols' must decrease between generations")
         }
     } else {
         checkInput(ngen, c("vector", "numeric"), 1, int = TRUE, gt = 0)
         checkInput(ptols, c("vector", "numeric"), 1, gt = 0, lt = 1)
-        tols <- apply(abs(t(x$output[[length(x$output)]]) - x$data), 1, stats::quantile, probs = ptols)
+        tols <- bisectTols(x$output[[length(x$output)]], x$data, x$tols[nrow(x$tols), ], ptols, ptollim = 0.1)
+        names(tols) <- colnames(x$tols)
         tols <- ifelse(tols < 0, 0, tols)
+        ## check against min tols
+        if(!is.null(x$mintols[1])) {
+            tols <- ifelse(tols < x$mintols, x$mintols, tols)
+        }
         if(all(tols == x$tols[nrow(x$tols), ])){
             stop("Tolerances same as previous generation at this 'ptol'")
         }
@@ -232,6 +262,7 @@ ABCSMC.ABCSMC <- function(x, tols = NULL, ptols = NULL, ngen = 1, parallel = FAL
         npart = nrow(x$pars[[1]]), 
         tols = tols, 
         ptols = ptols,
+        mintols = x$mintols,
         ngen = ngen,
         priors = x$priors, 
         func = x$func, 
@@ -255,6 +286,7 @@ ABCSMC.ABCSMC <- function(x, tols = NULL, ptols = NULL, ngen = 1, parallel = FAL
     x$ESS <- c(x$ESS, temp$ESS)
     x$accrate <- c(x$accrate, temp$accrate)
     x$tols <- rbind(x$tols, temp$tols)
+    x$ptols <- temp$ptols
     
     ## return new object
     x
@@ -263,7 +295,7 @@ ABCSMC.ABCSMC <- function(x, tols = NULL, ptols = NULL, ngen = 1, parallel = FAL
 #' @rdname ABCSMC
 #' @export
 
-ABCSMC.default <- function(x, priors, func, u, tols = NULL, ptols = NULL,
+ABCSMC.default <- function(x, priors, func, u, tols = NULL, ptols = NULL, mintols = NULL,
                            ngen = 1, npart = 100, parallel = FALSE, mc.cores = NA, ...) {
     
     ## check missing arguments
@@ -335,10 +367,19 @@ ABCSMC.default <- function(x, priors, func, u, tols = NULL, ptols = NULL,
     }
     if(nrow(tols) > 1){
         if(!all(apply(tols, 2, function(x) {
-            all(diff(x) < 0)
+            all(diff(x) <= 0)
         }))){
             stop("'tols' cannot increase")
         }
+        temp <- apply(tols, 2, function(x) {
+            diff(x)
+        })
+        if(is.null(nrow(temp))) {
+            temp <- matrix(temp, nrow = 1)
+        }
+        if(apply(temp, 1, function(x) !any(x < 0))){
+            stop("Some 'tols' must decrease between generations")
+        }
     }
     if(is.null(ptols[1])){
         ngen <- nrow(tols)
@@ -351,6 +392,18 @@ ABCSMC.default <- function(x, priors, func, u, tols = NULL, ptols = NULL,
         }
     }
     
+    if(!is.null(mintols[1])) {
+        checkInput(mintols, c("vector", "numeric"), gte = 0, length = ncol(tols))
+        checkInput(names(mintols), c("vector", "character"), ident = colnames(tols))
+        if(is.null(ptols[1])) {
+            if(any(apply(rbind(mintols, tols), 2, function(x) {
+                any(x[1] > x[-1])
+            }))) {
+                stop("'tols' can't be less than 'mintols'")
+            }
+        }
+    }
+    
     ## check function
     fargs <- formals(func)
     if(length(fargs) < 4){
@@ -463,13 +516,17 @@ ABCSMC.default <- function(x, priors, func, u, tols = NULL, ptols = NULL,
         runind <- T
         if(t != init){
             ## set tolerances if required
-            if(!is.null(ptols[1])){
-                tols[t, ] <- apply(abs(t(out[[t - 1]]) - data), 1, stats::quantile, probs = ptols)
+            if(!is.null(ptols[1])) {
+                tols[t, ] <- bisectTols(out[[t - 1]], data, tols[t - 1, ], ptols, ptollim = 0.1)
                 tols[t, ] <- ifelse(tols[t, ] < 0, 0, tols[t, ])
+                ## check against min tols
+                if(!is.null(mintols[1])) {
+                    tols[t, ] <- ifelse(tols[t, ] < mintols, mintols, tols[t, ])
+                }
                 if(all(tols[t, ] == tols[t - 1, ])){
                     tols <- tols[1:(t - 1), , drop = FALSE]
                     runind <- F
-                    warning("Tolerances same as previous generation, so now stopping algorithm.")
+                    message("Tolerances same as previous generation, so now stopping algorithm.")
                 }
             }
         }
@@ -479,7 +536,7 @@ ABCSMC.default <- function(x, priors, func, u, tols = NULL, ptols = NULL,
         
         ## run generation
         if(runind){
-            if(!parallel) {
+            if(!parallel | npart == 1) {
                 temp <- lapply(1:npart, runProp,
                     t = t, priors = priors, 
                     prevWeights = tempWeights, prevPars = tempPars, 
@@ -487,6 +544,8 @@ ABCSMC.default <- function(x, priors, func, u, tols = NULL, ptols = NULL,
                     u = u,
                     func = func, func_args = fargs)
             } else  {
+                ## set RNG generator to ensure reproducibility
+                RNGkind("L'Ecuyer-CMRG")
                 temp <- parallel::mclapply(1:npart, runProp,
                     t = t, priors = priors, 
                     prevWeights = tempWeights, prevPars = tempPars, 
@@ -494,6 +553,8 @@ ABCSMC.default <- function(x, priors, func, u, tols = NULL, ptols = NULL,
                     u = u,
                     func = func, func_args = fargs, 
                     mc.cores = mc.cores)
+                ## reset RNG generator
+                RNGkind("default")
             }
             
             ## extract relative components
@@ -537,7 +598,7 @@ ABCSMC.default <- function(x, priors, func, u, tols = NULL, ptols = NULL,
     
     ## output results
     output <- list(pars = pars, output = out, weights = weights, ESS = ESS, accrate = accrate,
-                   tols = tols, ptols = ptols, priors = orig_priors, data = data,
+                   tols = tols, ptols = ptols, mintols = mintols, priors = orig_priors, data = data,
                    func = func, u = u, addargs = args)
     class(output) <- "ABCSMC"
     output
